// =====================================================================================================================
//  fennec, a free and open source game engine
//  Copyright © 2025  Medusa Slockbower
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <https://www.gnu.org/licenses/>.
// =====================================================================================================================

///
/// \file type_traits.h
/// \brief get info about types at compile-time
///
///
/// \details
/// \author Medusa Slockbower
///
/// \copyright Copyright © 2025 Medusa Slockbower ([GPLv3](https://www.gnu.org/licenses/gpl-3.0.en.html))
///
///

#ifndef FENNEC_LANG_TYPE_TRAITS_H
#define FENNEC_LANG_TYPE_TRAITS_H

#include <fennec/lang/type_transforms.h>
#include <fennec/lang/detail/__type_traits.h>

namespace fennec
{

// fennec::is_void =====================================================================================================

///
///
/// \brief check if \p T is of type void
///
/// \details
/// \tparam T type to check
template<typename T> struct is_void
	: detail::__is_void<remove_cvr_t<T>>{};

///
///
/// \brief shorthand for ```is_void<T>::value```
/// \tparam T type to check
template<typename T> constexpr bool_t is_void_v
	= is_void<T>::value;



// fennec::is_bool =====================================================================================================

///
///
/// \brief check if \p T is of type bool
///
/// \details
/// \tparam T type to check
template<typename T> struct is_bool
	: detail::__is_bool<remove_cvr_t<T>>{};

///
///
/// \brief shorthand for ```is_bool<T>::value```
/// \tparam T type to check
template<typename T> constexpr bool_t is_bool_v
	= is_bool<T>::value;



// Integral Types ======================================================================================================

///
///
/// \brief check if \p T is of an integral
///
/// \details
/// \tparam T type to check
template<typename T> struct is_integral
	: detail::__is_integral<remove_cvr_t<T>> {};

///
///
/// \brief shorthand for ```is_integral<T>::value```
/// \tparam T type to check
template<typename T> constexpr bool_t is_integral_v
	= is_integral<T>::value;


///
///
/// \brief check if \p T is of a signed integral
///
/// \details
/// \tparam T type to check
template<typename T> struct is_signed
	: detail::__is_signed<remove_cvr_t<T>> {};

///
///
/// \brief shorthand for ```is_signed<T>::value```
/// \tparam T type to check
template<typename T> constexpr bool_t is_signed_v
	= is_signed<T>::value;


///
///
/// \brief check if \p T is of an unsigned integral
///
/// \details
/// \tparam T type to check
template<typename T> struct is_unsigned
	: detail::__is_unsigned<remove_cvr_t<T>> {};

///
///
/// \brief shorthand for ```is_unsigned<T>::value```
/// \tparam T type to check
template<typename T> constexpr bool_t is_unsigned_v
	= is_unsigned<T>::value;



// Floating Point Types ================================================================================================

///
///
/// \brief check if \p T is of a floating point type
///
/// \details
/// \tparam T type to check
template<typename T> struct is_floating_point
	: detail::__is_floating_point<remove_cvr_t<T>>{};

///
///
/// \brief shorthand for ```is_floating_point<T>::value```
/// \tparam T type to check
template<typename T> constexpr bool_t is_floating_point_v
	= is_floating_point<T> {};



// Arithmetic Types ====================================================================================================

///
///
/// \brief check if \p T is an arithmetic type
///
/// \details
/// \tparam T type to check
template<typename T> struct is_arithmetic
	: bool_constant<is_integral_v<T> or is_floating_point_v<T>>{};

///
///
/// \brief shorthand for ```is_arithmetic<T>::value```
/// \tparam T type to check
template<typename T> constexpr bool_t is_arithmetic_v
	= is_arithmetic<T>::value;

// fennec::is_same =====================================================================================================

///
///
/// \brief check if the two types are identical
///
/// \details
/// \tparam T type to check
template<typename T0, typename T1> struct is_same
	: false_type {};

// true case
template<typename T> struct is_same<T, T>
	: true_type {};

///
///
/// \brief shorthand for ```is_same<T0, T1>::value```
/// \tparam T type to check
template<typename T0, typename T1> constexpr bool_t is_same_v
	= is_same<T0, T1> {};

// fennec::can_convert =================================================================================================

///
/// \brief check if type `T0` can be converted `T1`
/// \tparam T0 First type
/// \tparam T1 Second type
template<typename T0, typename T1> struct can_convert
	: bool_constant<FENNEC_BUILTIN_CAN_CONVERT(T0, T1)> {};

///
/// \brief shorthand
/// \param T0 First type
/// \param T1 Second type
template<typename T0, typename T1> using can_convert_v
	= typename can_convert<T0, T1>::type;


// fennec::is_constructible ===============================================================================================

///
/// \brief Check if `ClassT` can be constructed with `ArgsT,` i.e. `ClassT(ArgsT...)`.
///        This may be read as "is `ClassT` constructible with `ArgsT`"
/// \tparam ClassT The class type to test
/// \tparam ArgsT The arguments for the specific constructor
template<typename ClassT, typename...ArgsT> struct is_constructible
	: bool_constant<FENNEC_BUILTIN_IS_CONSTRUCTIBLE(ClassT, ArgsT...)> {};

template<typename ClassT, typename...ArgsT> constexpr bool_t is_constructible_v
	= is_constructible<ClassT, ArgsT...>{};


// fennec::
	
//

}

#endif // FENNEC_LANG_TYPE_TRAITS_H
